/*
 *  curses.c -- minimalistic curses implementation
 *
 *  AUTHOR: Alain Knaff (alain@knaff.lu)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <sys/select.h>
#include "curses.h"

/* "Framebuffer". Has an associated chtype array and a width */
struct fbuf {
  chtype *ch;
  int cols;
  int iscopy;
};

/* Gets the adress of character at position (y,x) within buffer b */
static chtype *getAddr(struct fbuf *b, int y, int x)
{
  return b->ch + y * b->cols + x;
}

/* Allocates buffer of given size*/
static int allocFBuffer(struct fbuf *b, int lines, int cols)
{
  b->ch = calloc(lines * cols, sizeof(chtype));
  if(b->ch == NULL)
    return -1;
  b->cols = cols;
  b->iscopy = 0;
  return 0;
}

/* Sets up this buffer as a subbuffer */
static void subFBuffer(struct fbuf *dst, struct fbuf *src, int y, int x)
{
  dst->ch = getAddr(src, y, x);
  dst->cols = src->cols;
  dst->iscopy = 1;
}

/* Freesup this buffer */
static void freeFBuffer(struct fbuf *b)
{
  if(!b->iscopy)
    free(b->ch);
  b->ch = NULL;
}
     

/* Copies lines lines of cols width from srcy,srcx position in src buffer to
   dsty, dstx position in destination buffer. No bounds checking is done */
static void copyLines(struct fbuf *dst, int dsty, int dstx,
		      struct fbuf *src, int srcy, int srcx,
		      int lines, int cols)
{
  int line;
  chtype *dstptr = getAddr(dst, dsty, dstx);
  chtype *srcptr = getAddr(src, srcy, srcx);
  for(line = 0; line < lines;
      line++, dstptr += dst->cols, srcptr += src->cols)
    memcpy(dstptr, srcptr, cols * sizeof(chtype));
}

/* Same as above, but copies backwards (useful for scrolling) */
static void copyLinesDesc(struct fbuf *dst, int dsty, int dstx,
			  struct fbuf *src, int srcy, int srcx,
			  int lines, int cols)
{
  int line;
  chtype *dstptr = getAddr(dst, dsty+lines-1, dstx);
  chtype *srcptr = getAddr(src, srcy+lines-1, srcx);
  for(line = lines-1; line >= 0;
      line--, dstptr -= dst->cols, srcptr -= src->cols)
    memcpy(dstptr, srcptr, cols * sizeof(chtype));
}

/* Fills lines lines of cols width from srcy,srcx position in src buffer with
 * chtype */
static void fillLines(struct fbuf *dst, int dsty, int dstx,
		      chtype ch,
		      int lines, int cols)
{
  int line, col;
  chtype *dstptr = getAddr(dst, dsty, dstx);
  for(line = 0; line < lines; line++, dstptr += dst->cols)
    for(col = 0; col < cols; col++)
      dstptr[col] = ch;
}


static struct fbuf physical;
static struct fbuf virtual;

/* Cursor position for redraw */
static int scr_x=-1;
static int scr_y=-1;

/* current attributes */
static chtype scr_curAttr;

#define FG_BITS 8
#define BG_BITS 12

/* cursor position for showing cursor */
static int dsp_x;
static int dsp_y;

/* do we need to clear the main window? */
static int needClear=1;
static chtype clearColor;

struct WINDOW {
	struct fbuf fbuf;
	int begin_y,begin_x;
	int cols, lines;
	int y, x; /* Cursor position */
	chtype curAttr; /* Current attribute */
	int isSubWin; /* is a sub window of another window? */
	int clear;
};

static void physGotoYX(int y, int x) {
	if(y != scr_y || x != scr_x)
		printf("\e[%d;%dH", y+1, x+1);
}

static int updateColor(chtype attr, int bits, int colorBase, int notfirst)
{
	unsigned char shouldColor = (attr >> bits) & 7;
	unsigned char isColor = (scr_curAttr >> bits) & 7;

	if(shouldColor == isColor)
		return 0;

	if(notfirst)
		putc(';', stdout);

	printf("%d", colorBase + shouldColor);
	scr_curAttr &= ~(7 << bits);
	scr_curAttr |= shouldColor << bits;
	return 1;
}

static int updateBool(chtype attr, chtype mask,
		      int yesCode, int noCode, int notfirst)
{
	chtype should = attr & mask;
	chtype is = scr_curAttr & mask;

	if(should == is)
		return 0;

	if(notfirst)
		putc(';', stdout);
 
	printf("%d", should ? yesCode : noCode);
	scr_curAttr &= ~mask;
	scr_curAttr |= should ? mask : 0;
	return 1;
}

static void physSetAttr(chtype attr) {
	attr &= ~A_CHARTEXT & ~A_COLOR_VALID;
	if(attr != scr_curAttr) {
		int notfirst=0;
		
		printf("\e[");

		/* Undeline flag */
		notfirst |= updateBool(attr, A_UNDERLINE, 4, 24, notfirst);
		
		/* Reverse flag */
		notfirst |= updateBool(attr, A_REVERSE, 7, 27, notfirst);

		/* "Dim" flag */
		notfirst |= updateBool(attr, A_DIM, 2, 22, notfirst);

		/* Bold flag: */
		notfirst |= updateBool(attr, A_BOLD, 1, 22, notfirst);

		/*  Foreground color */
		notfirst |= updateColor(attr, FG_BITS, 30, notfirst);

		/* Background color */
		notfirst |= updateColor(attr, BG_BITS, 40, notfirst);
		
		fputc('m', stdout);				
	}
}

static void physClear(void) {
	physSetAttr(clearColor);
	printf("\e[2J");
	needClear = 0;
}

/* update one line */
static int doUpdateLine(int line, chtype *pline, chtype *vline, int cols, 
			int force) {
	int begin;
	int end;
	int col;

	if(force) {
		begin = 0;
		end = cols;
	} else {
		
		for(begin=0; begin < COLS; begin++) {
			if(pline[begin] != vline[begin])
				break;
		}
		if(begin == cols) {
			/* No change */
			return 0;
		}
		for(end=cols-1; end >= begin; end--) {
			if(pline[end] != vline[end])
				break;
		}
	}

	/* jump to begin */
	physGotoYX(line, begin);

	/* copy loop */
	for(col = begin; col <= end; col++) {
		physSetAttr(vline[col]);
		fputc(vline[col] & A_CHARTEXT, stdout);
		pline[col] = vline[col];
		scr_x++;
	}

	return -1;
}

static int doUpdate(int force) {
	/* TODO */
	int line;
	chtype *pline, *vline;
	int cols = physical.cols;

	/* do we need clear? */
	if(needClear) {
		fillLines(&physical, 0, 0,
			  (clearColor & ~A_CHARTEXT) | ' ',
			  LINES, physical.cols);
		physClear();
	}

	for(line=0, pline = getAddr(&physical,0,0), 
		    vline = getAddr(&virtual, 0,0);
	    line < LINES; 
	    line++, pline += physical.cols, vline += virtual.cols)
		doUpdateLine(line, pline, vline, cols, force);

	physGotoYX(dsp_y, dsp_x);

	fflush(stdout);
	return -1;
}

int wnoutrefresh(WINDOW *win) {
	copyLines(&virtual, win->begin_y, win->begin_x,
		  &win->fbuf, 0, 0,
		  win->lines, win->cols);
	dsp_y = win->y + win->begin_y;
	dsp_x = win->x + win->begin_x;
	if(win->clear) {
		needClear=1;
		clearColor=win->curAttr;
		win->clear=0;
	}
	return 0;
}

int wrefresh(WINDOW *win) {
	wnoutrefresh(win);
	doUpdate(0);
	return 0;
}

int wbkgdset(WINDOW *win, int attrs) {
	win->curAttr = attrs & ~A_CHARTEXT;
	return 0;
}


static WINDOW *_newwin(int nlines, int ncols, int begin_y, int begin_x) {
	WINDOW *win = calloc(1, sizeof(WINDOW));
	if(win == NULL)
		return NULL;
	win->lines = nlines;
	win->cols = ncols;
	win->begin_y = begin_y;
	win->begin_x = begin_x;
	return win;
}

WINDOW *newwin(int nlines, int ncols, int begin_y, int begin_x) {
	WINDOW *win = _newwin(nlines, ncols,  begin_y, begin_x);
	if(allocFBuffer(&win->fbuf, nlines, ncols) < 0) {
		free(win);
		return NULL;
	}
	return win;
}

WINDOW *subwin(WINDOW *win, int nlines, int ncols, int begin_y, int begin_x)
{
	WINDOW *ret = _newwin(nlines, ncols,  begin_y, begin_x);
	subFBuffer(&ret->fbuf, &win->fbuf,
		   begin_y - win->begin_y, begin_x - win->begin_x);
	return ret;
}

int delwin(WINDOW *win) {
	freeFBuffer(&win->fbuf);
	free(win);
	return 0;
}

int wmove(WINDOW *win, int y, int x)
{
	win->y = y;
	win->x = x;
	return 0;
}

int waddch(WINDOW *win, chtype ch)
{
	if(win->curAttr)
		ch = (ch & A_CHARTEXT) | win->curAttr;
	if(win->x < win->cols && win->y < win->lines) {
		*getAddr(&win->fbuf, win->y, win->x) = ch;
		win->x++;
	}
	return 0;
}

int werase(WINDOW *win)
{
	fillLines(&win->fbuf, 0, 0, win->curAttr|' ', win->lines, win->cols);
	win->clear = 1;
	return 0;
}

int waddstr(WINDOW *win, const char *str)
{
	while(*str) {
		waddch(win,
		       (((unsigned char)*str)  & A_CHARTEXT) | win->curAttr);
		str++;
	}
	return 0;
}

static bool isCharWaiting=FALSE;
static unsigned char charWaiting;

static int readCh(void) {
	if(isCharWaiting) {
		isCharWaiting=FALSE;
		return charWaiting;
	}
	if(read(0, &charWaiting, 1) < 0)
		return ERR;
	return charWaiting;
}

static int peekCh(void) {
	fd_set read_set;
	struct timeval tv;
	int n;
	FD_ZERO(&read_set);
	FD_SET(0, &read_set);
	tv.tv_sec = 0;
	tv.tv_usec = 3000;

	if(isCharWaiting)
		return charWaiting;
	n = select(1, &read_set, 0, 0, &tv);
	if(n == 0)
		return 0;
	if(read(0, &charWaiting, 1) < 0)
		return ERR;
	isCharWaiting=1;
	return charWaiting;
}

int wgetch(WINDOW *win __attribute__((unused)))
{
	int ch;
	while ((ch = readCh()) == '\014') {
		doUpdate(1);
	}
	if(ch == '\e' && peekCh() == '[') {
		/* TODO:  */

		readCh(); /* consume the [ */
		ch = readCh();
		switch(ch) {
		case 'A':
			return KEY_UP;
		case 'B':
			return KEY_DOWN;
		case 'C':
			return KEY_RIGHT;
		case 'D':
			return KEY_LEFT;
		case 'H':
			return KEY_HOME;
		case 'F':
			return KEY_END;
		case '1':
			readCh();
			return KEY_HOME;
		case '3':
			readCh();
			return KEY_DC;
		case '4':
			readCh();
			return KEY_END;
		case '5':
			readCh();
			return KEY_PPAGE;
		case '6':
			readCh();
			return KEY_NPAGE;
		default:
			fprintf(stderr, "Bad sequence ^[[%c\n", ch);
			return ch;
		}
	} else if(ch == '\r') {
		ch = '\n';
	}
	return ch;
}

WINDOW *stdscr=NULL;

void _getyx(WINDOW *win, int *y, int *x)
{
	*x = win->x;
	*y = win->y;
}

int wscrl(WINDOW *win, int n)
{
	int begin;
	if(n == 0)
		return 0;
	else if(n > 0) {
		/* scroll up */	
		if(n > win->lines)
			n = win->lines;
		copyLines(&win->fbuf, 0, 0,
			  &win->fbuf, n, 0,
			  win->lines - n, win->cols);
		begin = win->lines-n;
	} else {
		/* scroll down */
		n = -n;
		if(n > win->lines)
			n = win->lines;
		copyLinesDesc(&win->fbuf, n, 0,
			      &win->fbuf, 0, 0,
			      win->lines - n, win->cols);
		begin = 0;
	}

	/* fill remaining space with spaces */
	fillLines(&win->fbuf, begin, 0,
		  win->curAttr | 0x20,
		  n, win->cols);
	return -1;
}

#define NR_PAIRS 100
static chtype colorPairs[NR_PAIRS];

#define MK_COLOR(fg,bg) ((((fg)&7)<<FG_BITS)|(((bg)&7)<< BG_BITS)|A_COLOR_VALID)

int init_pair(int slot, int fg, int bg)
{
	if(slot >= NR_PAIRS || slot < 0)
		return -1;

	colorPairs[slot] =  MK_COLOR(fg, bg);
	return 0;
}

int COLOR_PAIR(int slot)
{
	if(slot >= NR_PAIRS || slot < 0)
		return -1;
	return colorPairs[slot];
}

/* Flash */
int flash(void)
{
	putc('\007', stdout);
	fflush(stdout);
	return 0;
}

static struct termios term_orig, my_term; // remember what the cooked mode was

WINDOW *initscr(void)
{
	if(stdscr != NULL)
		return stdscr;
	printf("\e[11m");
	allocFBuffer(&physical, LINES, COLS);
	allocFBuffer(&virtual, LINES, COLS);
	stdscr = newwin(LINES, COLS, 0, 0);
	tcgetattr(0, &term_orig);
	my_term = term_orig;
	
	return stdscr;
}

int endwin(void)
{
	tcsetattr(0, TCSANOW, &term_orig);
	printf("\e[0m\e[1;1H\e[2J");
	fflush(stdout);
	return 0;
}

chtype winch(WINDOW *win)
{
	return *getAddr(&win->fbuf, win->y, win->x);
}

int cbreak(void)
{
	my_term.c_lflag &= ~ICANON;

        my_term.c_iflag &= (~IXON & ~ICRNL);
        my_term.c_oflag &= (~ONLCR);
        my_term.c_cc[VMIN] = 1;
        my_term.c_cc[VTIME] = 0;

	tcsetattr(0, TCSANOW, &my_term);
	return 0;
}

int noecho(void)
{
	my_term.c_lflag &= ~ECHO ;
	tcsetattr(0, TCSANOW, &my_term);
	return 0;
}
