/*
 *  menubox.c -- implements the menu box
 *
 *  AUTHOR: Savio Lam (lam836@cs.cuhk.hk)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include "dialog.h"
#include "udpc_lang.h"
#include "udpc_dialog.h"

static int menu_width, tag_x, item_x;
static unsigned int tag_width, item_width;

/*
 * Print menu item
 */
static void
print_item (WINDOW * win, const char *tag, const char *item,
		int choice, int selected)
{
    int i;

    /* Clear 'residue' of last item */
    wbkgdset (win, menubox_attr);
    wmove (win, choice, 0);
    for (i = 0; i < menu_width; i++)
	waddch (win, ' ');
    wmove (win, choice, tag_x);
    wbkgdset (win, selected ? tag_key_selected_attr : tag_key_attr);
    waddch (win, tag[0]);
    wbkgdset (win, selected ? tag_selected_attr : tag_attr);
    waddstr (win, tag + 1);
    wmove (win, choice, item_x);
    wbkgdset (win, selected ? item_selected_attr : item_attr);
    waddstr (win, item);
}

/* Draw arrows. The item_no parameter should actually be
 * item_no - max_choice */
static void drawArrows(WINDOW *dialog, int item_no, int scrolly,
		       int box_y, int box_x,
		       int box_height,
		       int cur_y, int cur_x)
{
    /* print the up/down arrows */
    wmove (dialog, box_y, box_x + tag_x + 1);

    wbkgdset (dialog, scrolly ? uarrow_attr : menubox_attr);
    waddch (dialog, scrolly ? ACS_UARROW : ACS_HLINE);
    wmove (dialog, box_y, box_x + tag_x + 2);
    waddch (dialog, scrolly ? '(' : ACS_HLINE);
    wmove (dialog, box_y, box_x + tag_x + 3);
    waddch (dialog, scrolly ? '-' : ACS_HLINE);
    wmove (dialog, box_y, box_x + tag_x + 4);
    waddch (dialog, scrolly ? ')' : ACS_HLINE);

    wbkgdset (dialog, scrolly < item_no ? darrow_attr : menubox_border_attr);
    wmove (dialog, box_y + box_height + 1,  box_x + tag_x + 1);
    waddch (dialog, scrolly < item_no ?  ACS_DARROW : ACS_HLINE);
    wmove (dialog, box_y + box_height + 1,  box_x + tag_x + 2);
    waddch (dialog, scrolly < item_no ?  '(' : ACS_HLINE);
    wmove (dialog, box_y + box_height + 1,  box_x + tag_x + 3);
    waddch (dialog, scrolly < item_no ?  '+' : ACS_HLINE);
    wmove (dialog, box_y + box_height + 1,  box_x + tag_x + 4);
    waddch (dialog, scrolly < item_no ?  ')' : ACS_HLINE);

    if(cur_y != -1) {
	wmove (dialog, cur_y, cur_x);
	wrefresh (dialog);
    }
}

static int quickJump(WINDOW *menu, const char * const *items,
		     int max_choice, int *scrolly, int target, int *scrolled)
{
    int i;
    int choice;
    int alwaysInitialize = 0;
    int oldScrolly;
    if(*scrolly == -1) {
	alwaysInitialize=1;
	*scrolly=0;
    }
    oldScrolly=*scrolly;

    if(target > *scrolly + max_choice - 1) {
	*scrolly = target - max_choice+1;
	choice = max_choice-1;
    } else if(target < *scrolly) {
	*scrolly =target;
	choice=0;
    } else {
	choice = target - *scrolly;
	if(!alwaysInitialize)
	    return choice;
    }

    /* Print the menu */
    for (i = 0; i < max_choice; i++)
	print_item (menu, items[(*scrolly+i) * 2], items[(*scrolly+i) * 2 + 1],
		    i, i == choice);
    wnoutrefresh (menu);
    if(scrolled) {
	if(*scrolly != oldScrolly)
	    *scrolled=1;
	else
	    *scrolled=0;
    }
    return choice;
}

static int findItemByInitial(int item_no, const char * const *items, 
			     int start, char key)
{
    int i;
    key = toupper(key);

	/* start searching after current */
    for (i = start+1; i < item_no; i++)
	if (key == toupper (items[i * 2][0]))
	    return i;

    /* then loop back to beginning */
    for (i = 0; i < start+1; i++)
	if (key == toupper (items[i * 2][0]))
	    return i;
    return -1; /* not found */
}


/*
 * Display a menu for choosing among a number of options
 */
int
dialog_menu (const char *title, const char *prompt, int height, int width,
	     int menu_height, int item_no, const char * const * items,
	     int preselect)
{
    int i, x, y, cur_x, cur_y, box_x, box_y, box_height;
    int key = 0, button = 0, choice = 0, scrolly = -1, max_choice;
    WINDOW *dialog, *menu;

    menu_height = MIN (menu_height, item_no);
    max_choice = MIN (menu_height, item_no);
    height = MIN (height, menu_height + 7);

    /* center dialog box on screen */
    x = (COLS - width) / 2;
    y = (LINES - height) / 2;

#ifdef HAVE_NCURSES
    if (use_shadow)
	draw_shadow (stdscr, y, x, height, width);
#endif
    dialog = newwin (height, width, y, x);
    keypad (dialog, TRUE);

    draw_box (dialog, 0, 0, height, width, dialog_attr, border_attr);
    wbkgdset (dialog, border_attr);
    wmove (dialog, height - 3, 0);
    waddch (dialog, ACS_LTEE);
    for (i = 0; i < width - 2; i++)
	waddch (dialog, ACS_HLINE);
    wbkgdset (dialog, dialog_attr);
    waddch (dialog, ACS_RTEE);
    wmove (dialog, height - 2, 1);
    for (i = 0; i < width - 2; i++)
	waddch (dialog, ' ');

    if (title != NULL) {
	wbkgdset (dialog, title_attr);
	wmove (dialog, 0, (width - strlen (title)) / 2 - 1);
	waddch (dialog, ' ');
	waddstr (dialog, title);
	waddch (dialog, ' ');
    }
    wbkgdset (dialog, dialog_attr);
    print_autowrap (dialog, prompt, width - 2, 1, 3);

    menu_width = width - 6;
    getyx (dialog, cur_y, cur_x);
    box_y = cur_y + 1;
    box_x = (width - menu_width) / 2 - 1;
    box_height = menu_height;
    if( box_height >= height -6 )
	box_height = height -6;
    /* create new window for the menu */
    menu = subwin (dialog, box_height, menu_width, y + box_y + 1,

		   x + box_x + 1);
    keypad (menu, TRUE);

    /* draw a box around the menu items */
    draw_box (dialog, box_y, box_x, box_height + 2, menu_width + 2,
	      menubox_border_attr, menubox_attr);

    tag_width  = 0;
    item_width = 0;
    /* Find length of longest item in order to center menu */
    for (i = 0; i < item_no; i++) {
        tag_width  = MAX (tag_width, strlen (items[i * 2]));
        item_width = MAX (item_width, strlen (items[i * 2 + 1]));
    }
    tag_x = (menu_width - (tag_width + item_width + 2)) / 2;
    item_x = tag_x + tag_width + 2;

    /* Print the menu */
    choice=quickJump(menu, items, max_choice, &scrolly, preselect, NULL);

    drawArrows(dialog, item_no-max_choice, 
	       scrolly, box_y, box_x, box_height, -1, -1);

    x = width / 2 - 11;
    y = height - 2;
    print_button (dialog, cancelBuffer, y, x + 14, FALSE);
    print_button (dialog, okBuffer, y, x, TRUE);
    wrefresh (dialog);

    while (key != ESC) {
	key = wgetch (dialog);
	/* Check if key pressed matches first character of any
	   item tag in menu */
	i = max_choice;
	if((key >= 'A' && key <= 'Z') ||
	   (key >= 'a' && key <= 'z')) {
	    i = findItemByInitial(item_no, items, scrolly+choice, key);
	    if(i != -1) {
		int scrolled;
		getyx (dialog, cur_y, cur_x);		
		i=quickJump(menu, items, max_choice, &scrolly, i, &scrolled);
		if(scrolled)
		    drawArrows(dialog, item_no-max_choice, scrolly, 
			       box_y, box_x, box_height, cur_y, cur_x);
	    } else {
		i = max_choice; /* not found, mark invalid */
	    }
	}

	if (i < max_choice ||
	    (key >= '1' && key <= MIN ('9', '0' + max_choice)) ||
	    key == KEY_UP || key == KEY_DOWN || key == '-' ||
	    key == KEY_NPAGE || key == KEY_PPAGE ||
	    key == '+' ) {
	    if (key >= '1' && key <= MIN ('9', '0' + max_choice))
		i = key - '1';
	    else if(key == KEY_PPAGE) {
		int scrolled;
		getyx (dialog, cur_y, cur_x);
		if(!choice) {
		    if(scrolly < max_choice)
			i=0;
		    else
			i=scrolly-max_choice;
		    i=quickJump(menu, items,max_choice, &scrolly,i,&scrolled);
		    if(scrolled)
			drawArrows(dialog, item_no-max_choice, scrolly, 
				   box_y, box_x, box_height, cur_y, cur_x);
		} else {
		    i=0;
		}
	    }
	    else if(key == KEY_NPAGE) {
		if(choice<max_choice-1) {
		    if(max_choice-1+scrolly>item_no)
			i=item_no-scrolly;
		    else
			i=max_choice-1;
		} else {
		    int scrolled;
		    getyx (dialog, cur_y, cur_x);
		    i=scrolly+max_choice+max_choice-1;
		    if(i>item_no)
			i=item_no-1;
		    i=quickJump(menu, items,max_choice, &scrolly,i,&scrolled);
		    if(scrolled)
			drawArrows(dialog, item_no-max_choice, scrolly, 
				   box_y, box_x, box_height, cur_y, cur_x);
		}
	    }
	    else if (key == KEY_UP || key == '-') {
		if (!choice) {
		    if (scrolly) {

			/* Scroll menu down */
			getyx (dialog, cur_y, cur_x);
			if (box_height > 1) {
			    /* De-highlight current first item */
			    print_item (menu, items[scrolly * 2],
					items[scrolly * 2 + 1], 0, FALSE);
			    scrollok (menu, TRUE);
			    wscrl (menu, -1);
			    scrollok (menu, FALSE);
			}
			scrolly--;
			print_item (menu, items[scrolly * 2],
				    items[scrolly * 2 + 1], 0, TRUE);
			wnoutrefresh (menu);

			/* print the up/down arrows */
			drawArrows(dialog, item_no-max_choice, scrolly, 
				   box_y, box_x, box_height, cur_y, cur_x);
		    }
		    continue;	/* wait for another key press */
		} else
		    i = choice - 1;
	    } else if (key == KEY_DOWN || key == '+') {
		if (choice == max_choice - 1) {
		    if (scrolly + choice < item_no - 1) {
			/* Scroll menu up */
			getyx (dialog, cur_y, cur_x);
			if (box_height > 1) {
			    /* De-highlight current last item */
			    print_item (menu, items[(scrolly + max_choice - 1)
				       * 2], items[(scrolly + max_choice - 1)
					   * 2 + 1], max_choice - 1, FALSE);
			    scrollok (menu, TRUE);
			    scroll (menu);
			    scrollok (menu, FALSE);
			}
			scrolly++;
			print_item (menu, items[(scrolly + max_choice - 1) * 2],
				    items[(scrolly + max_choice - 1) * 2 + 1],
				    max_choice - 1, TRUE);
			wnoutrefresh (menu);

			/* print the up/down arrows */
			drawArrows(dialog, item_no-max_choice, scrolly, 
				   box_y, box_x, box_height, cur_y, cur_x);
		    }
		    continue;	/* wait for another key press */
		} else
		    i = choice + 1;
	    }
	    if (i != choice) {
		/* De-highlight current item */
		getyx (dialog, cur_y, cur_x);	/* Save cursor position */
		print_item (menu, items[(scrolly + choice) * 2],
			    items[(scrolly + choice) * 2 + 1], choice, FALSE);

		/* Highlight new item */
		choice = i;
		print_item (menu, items[(scrolly + choice) * 2],
			    items[(scrolly + choice) * 2 + 1], choice, TRUE);
		wnoutrefresh (menu);
		wmove (dialog, cur_y, cur_x);
		wrefresh (dialog);
	    }
	    continue;		/* wait for another key press */
	}
	switch (key) {
	case ' ':
	case TAB:
	case KEY_LEFT:
	case KEY_RIGHT:
	    if (!button) {
		button = 1;	/* Indicates "Cancel" button is selected */
		print_button (dialog, okBuffer, y, x, FALSE);
		print_button (dialog, cancelBuffer, y, x + 14, TRUE);
	    } else {
		button = 0;	/* Indicates "OK" button is selected */
		print_button (dialog, cancelBuffer, y, x + 14, FALSE);
		print_button (dialog, okBuffer, y, x, TRUE);
	    }
	    wrefresh (dialog);
	    break;
	case '\n':
	    delwin (dialog);
	    return (button? -2 : (scrolly + choice));
	case ESC:
	    break;
	}
    }

    delwin (dialog);
    return -1;			/* ESC pressed */
}
