/*
 *  udpc_conffile.c -- udpcast options dialog for udpcast boot system
 *
 *  AUTHOR: Alain Knaff (alain@knaff.lu)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <errno.h>
#include <dirent.h>
#include <ftw.h>
#include <linux/hdreg.h>
#include <sys/ioctl.h>

#include "udpc_dialog.h"
#include "libbb.h"
#include "dialog.h"
#include "udpc_lang.h"
#include "udpc_conffile.h"

/* ===============================================================
 * Setup of Udpcast itself
 * =============================================================== */


/* Input of udpcast port parameter */
static int configurePort(int needInput, int dir UNUSED, void *data UNUSED)
{
  char buffer[40]="";
  int port=0;

  if(!udpc_config.portIsSet)
    needInput=1;
  if(udpc_config.portIsSet) {
    sprintf(buffer, "%d", udpc_config.port);
    port = udpc_config.port;
  } else {
    strcpy(buffer, "9000");
  }
  while(1) {
    if(!needInput) {
      if(port % 2) {
	udpc_alert(MSG(TXT_UDPCAST_PORT_MUST_BE_EVEN));
	dialog_clear();
      } else {
	udpc_config.port = port;
	udpc_config.portIsSet = 1;
	dialog_clear();
	return DIALOG_FORWARD;
      }
    }
    if(dialog_inputbox(MSG(TXT_TITLE_UDPCAST), MSG(TXT_UDPCAST_PORT),
		       10, 40, buffer)) {
      dialog_clear();
      return DIALOG_BACK;
    }
    if(dialog_input_result[0]) {
      needInput=0;
      port = strtol(dialog_input_result,0,0);
    } else {
      needInput=1;
    }
  }
}


#define MAX_DISKS 200

static unsigned int diskIdx;
static struct diskName {
  char name[11];
} *diskNames=NULL;

static const char **diskPtr;


static int registerDiskCb(const char *path, const struct stat *sb, int flag)
{
  struct stat buf;
  (void) flag;
  if(stat(path, &buf) >= 0 &&
     S_ISBLK(buf.st_mode) &&
     major(sb->st_rdev) != 1 && 
     major(sb->st_rdev) != 2) {
    diskPtr[2*diskIdx]=" ";
    diskPtr[2*diskIdx+1]=strdup(path);    
    diskIdx++;
  }
  return 0;
}

static int makeDiskNodes(void)
{
  if(diskPtr == NULL) {
    diskPtr = (const char **) xmalloc(2*(MAX_DISKS+1)*sizeof(char*));
    diskNames = 
      (struct diskName *) xmalloc(MAX_DISKS * sizeof(struct diskName));
    diskIdx=0;
    if(diskIdx == 0) {
      udpc_makeDevices();
      ftw("/dev", registerDiskCb, 10);
    }
    diskPtr[2*diskIdx] = " ";
    diskPtr[2*diskIdx+1] = "OTHER";
  }
  return 0;
}

/* Input of disk name */
static int inputDiskName(int needInput)
{
  makeDiskNodes();
  if(!udpc_config.disk[0])
    needInput=1;

  while(1) {
    int r;
    if(!needInput && udpc_config.disk[0]) {
      if(!udpc_fileExists(udpc_config.disk)) {
	udpc_alert(MSG(TXT_DISK_NOT_EXIST), udpc_config.disk, 
		   strerror(errno));
	dialog_clear();
      } else {
	dialog_clear();
	return 0;
      }
    }

    r=dialog_inputbox(MSG(TXT_TITLE_UDPCAST), MSG(TXT_UDPCAST_PARTITION),
		      10, 40, udpc_config.disk);
    if(r) {
      dialog_clear();
      return -1;
    }
    nt_strncpy(udpc_config.disk, dialog_input_result, DISKNAMESIZE);
    needInput = 0;
  }
}


static int choseDisk(int needInput)
{
  makeDiskNodes();
  if(!udpc_config.disk[0])
    needInput=1;
  if(!needInput && udpc_fileExists(udpc_config.disk))
    return 0;
  return udpc_choseFromList(needInput,
			    TXT_TITLE_UDPCAST, MSG(TXT_UDPCAST_PARTITION), 
			    23, 45, 16, 
			    diskIdx+1, diskPtr, 1,
			    udpc_config.disk, DISKNAMESIZE,
			    inputDiskName);
}


static int enterUdpcastMode(int forceManual,
			    int dir UNUSED, void *data UNUSED)
{
  int r;
  const char *choices[12] = { "rcv", 0, "snd", 0 };
  choices[1] = MSG(TXT_START_RECEIVER);
  choices[3] = MSG(TXT_START_SENDER);

  r=udpc_choseFromList(forceManual,
		       TXT_TITLE_UDPCAST, MSG(TXT_UDPCAST_DIRECTION),
		       14, 40, 2, 
		       2,choices,
		       0, udpc_config.mode, sizeof(udpc_config.mode),
		       NULL);
  if(r < 0)
    return DIALOG_BACK;
  udpc_config.modeIdx = r;
  return DIALOG_FORWARD;
}

static int enterUdpcastCompressionMode(int forceManual,
				       int dir UNUSED, void *data UNUSED)
{
  int r;
  const char *choices[6] = { "none", 0, "gz", 0, "lzop", 0 };
  choices[1] = MSG(TXT_NO_COMPRESS);
  choices[3] = MSG(TXT_GZ);
  choices[5] = MSG(TXT_LZOP);

  r=udpc_choseFromList(forceManual,
		       TXT_TITLE_UDPCAST, MSG(TXT_UDPCAST_COMPRESSION),
		       14, 40, 3, 
		       3,choices,
		       0, udpc_config.compr, sizeof(udpc_config.compr),
		       NULL);
  if(r < 0)
    return DIALOG_BACK;
  udpc_config.comprIdx = r;
  return DIALOG_FORWARD;
}

/* Input of udpcast parameters */
static int enterUdpcastParams(int forceManual,
			      int dir UNUSED, void *data UNUSED)
{
  int r;
  if(udpc_config.udpcParamIsSet && !forceManual)
    return DIALOG_FORWARD;
  r=dialog_inputbox(MSG(TXT_TITLE_UDPCAST), MSG(TXT_UDPCAST_PARAMS),
		    10, 40, udpc_config.udpcParam);
  dialog_clear();
  if(r)
    return DIALOG_BACK;
  nt_strncpy(udpc_config.udpcParam, dialog_input_result, UDPCPARAMSIZE);
  udpc_config.udpcParamIsSet=1;
  return DIALOG_FORWARD;
}

static int configureDisk(void)
{
  char *disk = strdup(udpc_config.disk);
  int i,d;
  
  for(i=strlen(disk)-1; i >= 0 && disk[i] >= '0' && disk[i] <= '9'; i--) {    
  }
  disk[i+1]='\0';
  d = open(disk, O_RDONLY);
  if(d < 0) {
    /* Do not raise an error => this may simply mean that this is not a
       standard SCSI or IDE disk, and its partitions are simply following a
       different naming convention */
    free(disk);
    return 0;
  }

  /* FIXME: should only do this if it is an IDE disk */
  ioctl(d, HDIO_SET_DMA, 1);
  ioctl(d, HDIO_SET_UNMASKINTR, 1);
  ioctl(d, HDIO_SET_32BIT, 1);
  ioctl(d, HDIO_SET_MULTCOUNT, 16);
  close(d);

  free(disk);
  return 0;
}

static int askForDisk(int forceManual, int dir UNUSED, void *data UNUSED);

static configFn_t menu[] = {
  configurePort,
  askForDisk,
  enterUdpcastParams,
  enterUdpcastCompressionMode,
  enterUdpcastMode,
  NULL
};


int udpc_configureUdpcast(int forceManual, int dir, void *data UNUSED)
{
  if(dir == DIALOG_FORWARD) {
    if(diskPtr != NULL) {
      free(diskPtr);
      diskPtr = NULL;
    }
    if(diskNames != NULL) {
      free(diskNames);
      diskNames = NULL;
    }    
  }
  if(udpc_config.automatic != 1)
    forceManual=1;
  return dialogSeries(menu, forceManual, dir, NULL);
}

static int askForDisk(int forceManual, int dir UNUSED, void *data UNUSED) {
  if(choseDisk(forceManual) < 0 || configureDisk() < 0)
    return DIALOG_BACK;
  else
    return DIALOG_FORWARD;
}
